// $Id: BasicAgent.cpp,v 1.6 2003/07/05 18:44:35 darseq Exp $
// copyright, s.post & cp

#include "BasicAgent.h"
#include "behaviour/behaviouridle.h"
#include "behaviour/behaviour.h"
#include "SituationModel.h"
#include "WorldModel.h"
#include "Summary.h"

BasicAgent::BasicAgent(int type) : BasicActor(type) {
    mLastPath = NULL;
}
//----------------------------------------------------------------------------

BasicAgent::~BasicAgent() {
    if (mLastPath) {
        delete mLastPath;
    }
}
//----------------------------------------------------------------------------

void BasicAgent::init() {
    BasicActor::init();
    mMoves = mActions = 0;
    mTotalThinkTime=0;

    phase = AA_SAY;
    center_msgs = MSGS_C_CA-1; // expect an CA_READY message
}
//----------------------------------------------------------------------------

void BasicAgent::update() {
    UpdateSenseTimes();
    // Update the summary
//    model()->GetSummary()->Update(model()); <- happens in communicateWorld
    communicateWorld();
}
//----------------------------------------------------------------------------

void BasicAgent::UpdateSenseTimes()
{
    // update sensed time of objects we walked over
    if (mLastPath)
    {
        for(int i=0;mLastPath[i];i++)
        {
            model()->SetSenseTime(mLastPath[i],m_time);
            if (mLastPath[i]==Position())
                break;
        }
        delete mLastPath;
        mLastPath=0;
    }
    // update sensed time of everything in visual range of my current position
    for(Objects::const_iterator OI=model()->objects().begin();OI!=model()->objects().end();OI++) {
        if (DistanceBetween(Position(),*OI) < config.vision()) {
            model()->SetSenseTime(Position(),m_time);
        }
    }
}
//----------------------------------------------------------------------------

void BasicAgent::think() {
    int timestart,timeend;
    timestart = timer();
    Debug(spam_spam,"%d (%ld) Thinking\n",agentName(),m_id);

    // throw away my previous thinking process
    if (mLastTree)
    {
        delete mLastTree;
        // behaviour is deleted and illegal now
        mLastBehaviour = NULL;
    }
    // Get a situational model of the real world to base my thinking on
    SituationModel *Situation,*NewSituation;
    unsigned long timeextract=timer();
    mLastTree=Situation=model()->ExtractCurrentSituation(this);
    Debug(spam_performance,"\Extract %ldms\n",timer()-timeextract);

    TeamMember* Actor;
    Behaviour *BestBehaviour;
    float best_heuristic=-10000.0f,heuristic;

    // for every actor that is not busy
    vector<float> heuristics;  // keeps heuristic score, cHeursticFail means the corresponding behaviour is predicted to fail
    const float cHeursticFail=-MAXFLOAT;

    Debug(/*spam_performance|*/spam_behaviour_selection,"Thinking with %d actors\n",Situation->Actors());

    while ((Actor=Situation->NextActor()))
    {
        // get all possible behaviours based on the situation (team, task and map)
        unsigned long timegenerate=timer();
        Situation->GenerateBehaviour(Actor);
        const Behaviours &behaviours=Situation->GetBehaviours();
        Debug(spam_performance,"\tGenerate behaviour %ldms\n",timer()-timegenerate);
        
        // calculate all heuristic scores and reset predicted flags
        Debug(/*spam_performance|*/spam_behaviour_selection,"Selecting from %d behaviours:\n",behaviours.size());

        long timeheuristic=timer();
        heuristics.resize(behaviours.size());
        int bi=0;
        for(Behaviours::const_iterator I=behaviours.begin();I!=behaviours.end();I++,bi++)
        {
            heuristics[bi]=(*I)->Heuristic(Situation,Actor);
            Debug(spam_behaviour_selection,"%s (%f)\t",(*I)->ToString().c_str(),heuristics[bi]);

        }
        Debug(spam_behaviour_selection,"\n");
        Debug(spam_performance,"\tHeuristics %ldms\n",timer()-timeheuristic);

        // find the most promising possible behaviour
        do
        {
            // find the most promising behaviour
            BestBehaviour=0;
            int bi=0,bestBi=0;
            for(Behaviours::const_iterator I=behaviours.begin();I!=behaviours.end();I++,bi++)
            {
                if (heuristics[bi]!=cHeursticFail)
                {
                    heuristic=heuristics[bi];
                    if ((!BestBehaviour) || (heuristic>best_heuristic))
                    {
                        bestBi=bi;
                        BestBehaviour=*I;
                        best_heuristic = heuristic;
                    }
                }
            }
            if (!BestBehaviour)
            {   // no behaviour. this actor will idle
                Debug(spam_behaviour_selection,"No behaviour possible: Idle\n");
                BestBehaviour=new BehaviourIdle();
            }
            Debug(spam_behaviour_selection,"Predicting %s\n",BestBehaviour->ToString().c_str());
            // predict what the results would be for acting out this behaviour
            // and use it as the next iteration of the thinking process
            unsigned long timepredict=timer();
            NewSituation=Situation->Predict(Actor,BestBehaviour);
            Debug(spam_performance,"\tPredict %s (%ldms)\n",BestBehaviour->ToString().c_str(),timer()-timepredict);

            if (!NewSituation)
            {
                Debug(spam_behaviour_selection,"Behaviour %s predicted to fail\n",BestBehaviour->ToString().c_str());
                heuristics[bestBi]=cHeursticFail;
            }
        } while(!NewSituation);

        Situation=NewSituation;
                        
        // TODO Situation=CurrentSituation->SelectMostPromisingNode();
    }
    // find out what I actually figured I would do
    Actor=Situation->GetTeammember(Self());
    mLastBehaviour=Actor->Behaviour();
    Debug(spam_behaviour_selection,"Selected %s\n",mLastBehaviour->ToString().c_str());

    timeend=timer();
    Debug(spam_performance,"Thinking of %s done. %ld\t time=%ld\n",agentName(),ID(),timeend-timestart);
    mTotalThinkTime+=(timeend-timestart);

    // and do it
    mLastBehaviour->Execute(this);
}
//----------------------------------------------------------------------------

void BasicAgent::Move(Path* path)
{
    mMoves++;

    if (mLastPath)
    {
        delete mLastPath;
    }

    Path P=*path;
    P.Optimize();
    assert(P.Consistency());
    mLastPath=P.GetArray();

    if ((mLastPath[0]) && (mLastPath[0]==Position()))
    {
        if (mLastPath[1])
        {
            Debug(spam_output,"Output:[Agent %ld Time %ld Pos %ld] Moving with skip %s\n",
                m_id,m_time,Position()->id(),P.ToString().c_str());
            act_move(mLastPath+1);
        }
        else
        {
            Debug(spam_output,"Output:[Agent %ld Time %ld Pos %ld] Moving skipping own position\n",
                m_id,m_time,Position()->id(),P.ToString().c_str());
        }
    }
    else
    {
        Debug(spam_output,"Output:[Agent %ld Time %ld Pos %ld] Moving along %s\n",
                m_id,m_time,Position()->id(),P.ToString().c_str());
        act_move(mLastPath);
    }
}
//----------------------------------------------------------------------------

void BasicAgent::Idle() {
    Debug(spam_output,"Output:[Agent %ld Time %ld] Idle\n",m_id,m_time);
}
//----------------------------------------------------------------------------

void BasicAgent::PrintResults(FILE* file) {
    fprintf(file,"Agent %04ld\n",m_id);
    fprintf(file,"\tType = %s\n",agentName());
    fprintf(file,"\tFrame = %ld\n",m_time);
    fprintf(file,"\tThinkTime = %ld\n",mTotalThinkTime);
    fprintf(file,"\tTimes acted = %d\n",mActs);
    fprintf(file,"\tMoves = %d\n",mMoves);
    fprintf(file,"\tActions = %d\n",mActions);
    fprintf(file,"\tSectors = [%d,%d]\n",model()->GetSectorX(),model()->GetSectorY());
    fprintf(file,"\tModelBytes = %ld\n",model()->mSectorBytes+model()->mHighwayBytes+model()->mPathBytes);
    fprintf(file,"\tAnalyzeTime = (sector:%ld + paths:%ld + highways:%ld )%ld\n",model()->mSectorTime,model()->mPathTime,model()->mHighwayTime,model()->mSectorTime+model()->mPathTime+model()->mHighwayTime);
    fprintf(file,"\tPaths = %ld\n",model()->mInternalPaths);
    fprintf(file,"\tHighways = %ld\n",model()->mHighways);
}
//----------------------------------------------------------------------------

Eltlst * BasicAgent::destroyDuplicates(S32 id,char * c) {
    Eltlst * lst;
    if((lst = dynamic_cast<Eltlst *>(say_msgs.getElt(c)))) {
        for(Eltlst::iterator it = lst->begin();it != lst->end();it++) {
            Scalar * scalar; Eltlst * sublst = dynamic_cast<Eltlst *>(*it);
            if(sublst && (scalar = dynamic_cast<Scalar *>(sublst->front())) &&
                atoi(scalar->c_str()) == id) {
                delete *it; lst->erase(it); break;
            }
        }
    } else {lst = new Eltlst(); say_msgs.insertElt(c,lst);}
    return lst;
}
//----------------------------------------------------------------------------

void BasicAgent::sense(Object * object) {
    // buffer messages for aa_say
    switch(object->type()) {
        case TYPE_FIRE_BRIGADE:
        case TYPE_POLICE_FORCE:
        case TYPE_AMBULANCE_TEAM:
        case TYPE_CIVILIAN: {
            Eltlst * lst = destroyDuplicates(object->id(),"A");
            Eltlst * sublst = new Eltlst(); // ~Message releases Elts
            Humanoid * h = dynamic_cast<Humanoid *>(object);
            ASSERT(h); ASSERT(h->position());
            sublst->push_back(new Scalar(IntToString(h->id())));
            sublst->push_back(new Scalar(IntToString(h->position()->id())));
            sublst->push_back(new Scalar(IntToString(h->hp())));
            sublst->push_back(new Scalar(IntToString(h->damage())));
            sublst->push_back(new Scalar(IntToString(h->buriedness())));
            lst->push_back(sublst);
        }
        break;

        case TYPE_REFUGE:
        case TYPE_BUILDING: {
            Eltlst * lst = destroyDuplicates(object->id(),"B");
            Eltlst * sublst = new Eltlst(); // ~Message releases Elts
            Building * b = dynamic_cast<Building *>(object); ASSERT(b);
            sublst->push_back(new Scalar(IntToString(b->id())));
            sublst->push_back(new Scalar(IntToString(b->fieryness())));
            sublst->push_back(new Scalar(IntToString(b->brokenness())));
            lst->push_back(sublst);
        }
        break;

        case TYPE_ROAD: {
            Eltlst * lst = destroyDuplicates(object->id(),"R");
            Eltlst * sublst = new Eltlst(); // ~Message releases Elts
            Road * r = dynamic_cast<Road *>(object); ASSERT(r);
            sublst->push_back(new Scalar(IntToString(r->id())));
            sublst->push_back(new Scalar(IntToString(r->block())));
            sublst->push_back(new Scalar(IntToString(r->repairCost())));
            lst->push_back(sublst);
        }
        break;
        default: break;
//        case TYPE_NULL: case TYPE_RIVER_NODE: case TYPE_FIRE_STATION:
//        case TYPE_RIVER: case TYPE_POLICE_OFFICE: case TYPE_WORLD:
//        case TYPE_AMBULANCE_CENTER: case TYPE_NODE: case TYPE_CAR:
    }
}
//----------------------------------------------------------------------------

bool BasicAgent::ready2send() {
    // cycle    m_time      : SEND / KA_HEAR
    // sense -> m_time += 1 : KA_SENSE
    // cycle    m_time      : SEND / KA_HEAR
    if(m_time == m_last_send) {
        Debug(spam_agent_communication,"SEND: waiting for kernel cycle to "
            "complete\n");
        return false;
    }

    switch(phase) {
        case AC_EXCHANGE:
            // exchange: only exchange once, after that wait until
            // hear() changes phase to AA_SAY. m_time == m_last_hear:
            // make sure we enter a `clean' cycle before exchanging.
            if(exchange) {
                if(m_time == m_last_hear) {
                    Debug(spam_agent_communication,"EXCHANGE: waiting for "
                        "kernel cycle to complete\n");
                } else {
                    exchange = false; hear_msgs = 0; // clean cycle
                    return true;
                }
            }
        break;
        case AA_SAY: return true; break;// always ready to send say messages
        default: break;
    }
    return false;
}
//----------------------------------------------------------------------------

void BasicAgent::hear(Object * sender,const char * msg) {
    switch(phase) {
        case AC_EXCHANGE:
            if(anAgent(sender)) {
//                Debug(spam_agent_communication,"say/tell ignored\n");
            } else if(myCenter(sender)) {
                if(++hear_msgs > MSGS_C_A) { // MSGS_C_A == A_MAXMSGS
                    Debug(spam_agent_communication,"10SYNC FAILED!\n");
                } else {Exchange(sender,msg); m_last_hear = m_time;}
                // phase immediately becomes AA_SAY so we won't miss any
                // center messages. we can only leave AC_EXCHANGE when
                // the centers have seen AC_EXCHANGE agent send messages
                if(hear_msgs == MSGS_C_A) {center_msgs = 0; phase = AA_SAY;}
                if(exchange) { // make sure this agent has send
                    Debug(spam_agent_communication,"EXCHANGE FAILED!\n");
                }
            } else {Debug(spam_agent_communication,"Invalid sender!\n");}
        break;
        case AA_SAY:
            if(myCenter(sender)) {
                if(++center_msgs > MSGS_C_CA) {
                    Debug(spam_agent_communication,"11SYNC FAILED!\n");
                } else if(center_msgs == MSGS_C_CA) {
                    // ready2send should have done hear_msgs = 0,
                    // before other center messages are received
                    exchange = true; phase = AC_EXCHANGE;
                }
            } else if(anAgent(sender)) {
                if(m_time > m_last_hear) {hear_msgs = 0;}
                // listen to, but do not interpret your own messsages
                if(sender != Self()) {
                    if(++hear_msgs > A_MAXMSGS) {
//                        Debug(spam_agent_communication,"say/tell ignored\n");
                    } else {Agentmsg(sender,msg); m_last_hear = m_time;}
                }
            } else {Debug(spam_agent_communication,"Invalid sender!\n");}
        break;
        default: break;
    }
}
//----------------------------------------------------------------------------

void BasicAgent::communicateWorld() {
    if(ready2send() == false) {return;}

    switch(phase) {
        case AC_EXCHANGE: {
            string msg1 = "AC_EXCHANGE"; string msg2;
            string message = model()->GetSummary()->Update();
            if(message.length() <= MSGS_A_C*MSG_SIZE) {
                msg1 = message.substr(0,MSG_SIZE);
                if(message.length() > MSG_SIZE) {
                    msg2 = message.substr(MSG_SIZE);
                }
            }

            act_tell((char *) msg1.c_str());
//            act_tell((char *) msg2.c_str());
            m_last_exchange = m_last_send = m_time;
        }
        break;
        case AA_SAY: {
            // make sure centers cannot hear any say messages:
            // assume that we don't move before say arrives at the kernel
            model()->GetSummary()->Update();

            double radius = config.voice();
            S32 x = ((RealObject *) Position())->x();
            S32 y = ((RealObject *) Position())->y();
            const objects_t & centers = model()->Centers();
            objects_t::const_iterator CI = centers.begin();
            for(;CI!=centers.end();CI++) {
                double dx = (*CI)->x() - x;
                double dy = (*CI)->y() - y;
                if(dx*dx + dy*dy <= radius * radius) {
                    Debug(spam_agent_communication,
                        "AGENT %ld: keeping quiet\n",m_id);
                    say_msgs.clear();
                    return;
                }
            }

            Elt * elt; string msg1,msg2,msg3;
            if((elt = say_msgs.getElt("A"))) {
                msg1 = string("AA_SAY") + FIELD_DELIM + "A" +
                    STRDATA_SEP + elt->toString();
                if(msg1.length() <= MSG_SIZE) {
                    act_say((char *) msg1.c_str());
                }
            }
            if((elt = say_msgs.getElt("B"))) {
                msg2 = string("AA_SAY") + FIELD_DELIM + "B" +
                    STRDATA_SEP + elt->toString();
                if(msg2.length() <= MSG_SIZE) {
                    act_say((char *) msg2.c_str());
                }
            }

            if((elt = say_msgs.getElt("R"))) {
                msg3 = string("AA_SAY") + FIELD_DELIM + "R" +
                    STRDATA_SEP + elt->toString();
                if(msg3.length() <= MSG_SIZE) {
                    act_say((char *) msg3.c_str());
                }
            }
//            act_say((char *) msg4.c_str());

            m_last_send = m_time;
            say_msgs.clear();
        }
        break;
        default: break;
    }
}
//----------------------------------------------------------------------------

// update low level information
void BasicAgent::Agentmsg(Object * sender,const char * msg) {
    Debug(spam_agent_communication,"AGENTMSG [%ld],t=%d Hear: %ld,\"%s\"\n",
        m_id,m_time,sender->id(),msg);

    Message M(msg); Eltlst * lst;
    ASSERT(M.header == "AA_SAY");

    // A=((id,pos,hp,dmg,bury),...)
    if((lst = dynamic_cast<Eltlst *>(M.getElt("A")))) {
        Eltlst::iterator it1; Eltlst * sublst;
        for(it1 = lst->begin();it1 != lst->end();it1++) {
            if((sublst = dynamic_cast<Eltlst *>(*it1))) {
                Eltlst::iterator it2 = sublst->begin();
                int agent_id   = atoi(((Scalar *) *it2)->c_str());
                Humanoid * h =
                    dynamic_cast<Humanoid *>(model()->get(agent_id));
                ASSERT(h);
                int agent_pos  = atoi(((Scalar *) *(++it2))->c_str());
                h->setPosition(m_time,model()->get(agent_pos));
                int agent_hp   = atoi(((Scalar *) *(++it2))->c_str());
                h->setHp(m_time,agent_hp);
                int agent_dmg  = atoi(((Scalar *) *(++it2))->c_str());
                h->setDamage(m_time,agent_dmg);
                int agent_bury = atoi(((Scalar *) *(++it2))->c_str());
                h->setBuriedness(m_time,agent_bury);
            }
        }
    }

    // B=((id,fire,broken),...)
    if((lst = dynamic_cast<Eltlst *>(M.getElt("B")))) {
        Eltlst::iterator it1; Eltlst * sublst;
        for(it1 = lst->begin();it1 != lst->end();it1++) {
            if((sublst = dynamic_cast<Eltlst *>(*it1))) {
                Eltlst::iterator it2 = sublst->begin();
                int building_id     = atoi(((Scalar *) *it2)->c_str());
                Building * b =
                    dynamic_cast<Building *>(model()->get(building_id));
                ASSERT(b);
                int building_fire   = atoi(((Scalar *) *(++it2))->c_str());
                b->setFieryness(m_time,building_fire);
                int building_broken = atoi(((Scalar *) *(++it2))->c_str());
                b->setBrokenness(m_time,building_broken);
            }
        }
    }

    // R=((id,block,repair),...)
    if((lst = dynamic_cast<Eltlst *>(M.getElt("R")))) {
        Eltlst::iterator it1; Eltlst * sublst;
        for(it1 = lst->begin();it1 != lst->end();it1++) {
            if((sublst = dynamic_cast<Eltlst *>(*it1))) {
                Eltlst::iterator it2 = sublst->begin();
                int road_id     = atoi(((Scalar *) *it2)->c_str());
                Road * r =
                    dynamic_cast<Road *>(model()->get(road_id));
                ASSERT(r);
                int road_block  = atoi(((Scalar *) *(++it2))->c_str());
                r->setBlock(m_time,road_block);
                int road_repair = atoi(((Scalar *) *(++it2))->c_str());
                r->setRepairCost(m_time,road_repair);
            }
        }
    }
}
//----------------------------------------------------------------------------

// update high level information
void BasicAgent::Exchange(Object * sender,const char * msg) {
    Debug(spam_agent_communication,"AGENTEXC [%ld],t=%d Hear: %ld,\"%s\"\n",
        m_id,m_time,sender->id(),msg);
    switch(hear_msgs % MSGS_C_A)
    {
        case 1:
            model()->GetSummary()->arSectorA((const byte*)msg,m_last_exchange);
        break;
        case 2:
            model()->GetSummary()->arSectorB((const byte*)msg,m_last_exchange);
        break;
        case 3:
            model()->GetSummary()->arAgentA((const byte*)msg,m_last_exchange);
        break;
        case 0:
            model()->GetSummary()->arHighwayA((const byte*)msg,m_last_exchange);
        break;
    }
}
//----------------------------------------------------------------------------

bool BasicAgent::anAgent(Object * object) {
    // can also be yourself but not a civilian -> header != AA_SAY
    return dynamic_cast<Humanoid *>(object) &&
            (dynamic_cast<Civilian *>(object) == NULL);
}
//----------------------------------------------------------------------------

bool BasicAgent::myCenter(Object * object) {
    TypeId m_type = Self()->type(); // different from agent_type
    Building * center = dynamic_cast<Building *>(object);

    if(center) {
        switch(center->type()) {
            case TYPE_FIRE_STATION:
                return m_type == TYPE_FIRE_BRIGADE;
            break;
            case TYPE_POLICE_OFFICE:
                return m_type == TYPE_POLICE_FORCE;
            break;
            case TYPE_AMBULANCE_CENTER:
                return m_type == TYPE_AMBULANCE_TEAM;
            break;
        }
    }
    return false;
}
//----------------------------------------------------------------------------

void BasicAgent::act_say(char * msg) {
    BasicActor::act_say(msg);

Debug(spam_agent_communication,"AGENT [%ld] Say: \"%s\"\n",m_id,msg);
}
//----------------------------------------------------------------------------

void BasicAgent::act_tell(char * msg) {
    BasicActor::act_tell(msg);
    Debug(spam_agent_communication,"AGENT [%ld] Tell: \"%s\"\n",m_id,msg);
}
//----------------------------------------------------------------------------

