
#ifndef SITUATIONMODEL_H
#define SITUATIONMODEL_H

/*
 * Situation model is a model of the real world in the mind of an agent
 * it is used for coordination and behaviour selection
 * it is created by extracting a current situation model from
 * the worldmodel and iteratively by predicting the results
 * of the execution of the behaviour by the actor.
 */

#include <list>
#include <memory>
#include <adk.hxx>
#include "ImaginaryObject.h"
#include "behaviour/behaviour.h"

using namespace Rescue;

    class SelectionFunction
    {
    protected:
    public:
        virtual bool Consider(ImaginaryObject* aObject) const =0;
        virtual float Value(ImaginaryObject* aObject) const =0;
    };

class SituationModel : public list<class SituationModel*> 
{
friend class WorldModel;
friend class SituationForm;
friend class Behaviour;
protected:
    // information that doesn't change during thinking
    class WorldModel* mWorldModel;
    class Sector* mSector;
    class Task* mTask;
    class Team* mTeam;
    ImaginaryObjects *mOriginalObjects;
    class Rescue::Refuge* mClosestRefuge;
    
    // information that caused the change
    // the starting position
    class SituationModel* mParent;
    // the changer (overrides it's instance in mTeam)
    class TeamMember* mActor;
    
    // changed information
    // the time of the simulator
    long mTime;
    // objects that have changed due to actions
    ImaginaryObjects mOverrideObjects;
    

    // paths we're using while the paths from mSector aren't clear
    // size is Objects in sector count squared
    // NULL at places where we take the sector ones


    vector<class Path*> mInternalPaths;
    // routes we're using while the routes from mSector aren't clear
    // size is Sector count squared
    // NULL at places where we take the sector ones
    vector<class Route*> mRoutes;

    // changes for the future
    Behaviours mBehaviours;

    // internal use function
    void DoFindObjects(ImaginaryObjects& aOut,const SelectionFunction& aFn);
public: 
    SituationModel() 
    {
        mWorldModel=0;
        mSector=0;
        mTask=0;
        mTeam=0;
        mOriginalObjects=0;

        mClosestRefuge=0;
        mParent=0;
        mActor=0;
    }
    ~SituationModel();

    SituationModel(SituationModel& aOriginal) : list<class SituationModel*>()
    {
        // don't copy children. child should have its own children
        mWorldModel=aOriginal.mWorldModel;
        mSector=aOriginal.mSector;
        mTeam=aOriginal.mTeam;
        mTask=aOriginal.mTask;
        mClosestRefuge=aOriginal.mClosestRefuge;
        mOriginalObjects=aOriginal.mOriginalObjects;
        mActor=0;
        mParent=&aOriginal;
    }

    ImaginaryObject* Add(Rescue::Object* aObject);
    ImaginaryObject* Override(ImaginaryObject* aNewObject);
    
    // Gets the most recent state of an object
    ImaginaryObject* GetObject(const Object* aObject);
    ImaginaryBuilding* GetBuilding(const Object* aObject)
    {
        return dynamic_cast<ImaginaryBuilding*>(GetObject(aObject));
    }
    ImaginaryRoad* GetRoad(const Object* aObject)
    {
        return dynamic_cast<ImaginaryRoad*>(GetObject(aObject));
    }
    ImaginaryNode* GetNode(const Object* aObject)
    {
        return dynamic_cast<ImaginaryNode*>(GetObject(aObject));
    }
    ImaginaryHuman* GetHuman(const Object* aObject)
    {
        return dynamic_cast<ImaginaryHuman*>(GetObject(aObject));
    }
    // get all objects relevant to this situation for which aFn.Consider(O) returns true
    ImaginaryObjects FindObjects(const SelectionFunction& aFn);
    // get all objecst for which aFn.Consider(O) returns true so that the first
    // object in the container has the HIGHEST value for aFn.Value(O)
    ImaginaryObjects FindSortedObjects(const SelectionFunction& aFn);
    // Gets the most recent state of a teammember
    class TeamMember* GetTeammember(Object* aObject);
    // gets the entire teammember for iteratorion
    const class
    Team* GetTeam() { return mTeam; }
    
    // gets the best path in this situation
    class Path* FindPath(ImaginaryHuman* oTraveler,ImaginaryObject* oDestination,bool aTravel=true);

    // Get the next teammember that should act
    class TeamMember* NextActor();
    int Actors();
        
    // make behaviours according to this task, team, situation and first idle TeamMember
    void GenerateBehaviour(class TeamMember* aActor);
    const class Behaviours& GetBehaviours() { return mBehaviours; }
    
    // Predict the effect that the exectution of a behaviour by an actor will have
    // behaviour 0 means that actor will be idle
    class SituationModel* Predict(class TeamMember* aActor,class Behaviour* aBehaviour);
    // calculate the score of this situation
    float Score();

    // various extra getters
    bool InSector(class TeamMember* aAgent,class Sector* aSector);
    Rescue::Refuge* GetClosestRefuge() const { return mClosestRefuge; }
    class WorldModel* WorldModel() const { return mWorldModel; }
    // Debug functions
    void Draw(class ViewerWindow* aWindow);

    // Represent the situation as a short string
    string ToString() { return string("Situation"); }


    void SetTeam(class SummaryTeam* aTeam);
    const class Sector* GetSector() { return mSector; }
};

#endif
