#ifndef WIN_H
#define WIN_H 1

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string>
#include <vector>

using namespace std;

    enum view_enum
    {
        // clean view=15 sectorview=3599 sector errors=15777
        view_buildings=1<<0, // 1
        view_roads=1<<1, // 2
        view_nodes=1<<2, // 4
        view_agents=1<<3, // 8
        view_sense=1<<4, // 16
        view_browser=1<<5, // 32
        view_cursor=1<<6, // 64
        view_behaviour=1<<7, // 128
        view_tree=1<<8, // 256
        view_borders=1<<9, // 512
        view_nexi=1<<10, // 1024
        view_sectorcontent=1<<11, // 2048
        view_patherrors=1<<12, // 4096
        view_sectorerrors=1<<13, // 8192
        view_summary=1<<14, // 16384
        view_situation=1<<15, // 32768
        view_reserved1=1<<16 // 65536
    };
        
        
    class View
    {
    protected:
        double mScrollX,mScrollY,mZoom;
        int mViewX1,mViewY1,mViewX2,mViewY2;
        int mClipX1,mClipY1,mClipX2,mClipY2;
        bool mMirrorX,mMirrorY;
        bool mSquare;
        char* mName;        
		unsigned int mWidth, mHeight;
        void Recompute();
    public:
        View()
        {
            mScrollX=0;mScrollY=0;mZoom=1.0;
            mViewX1=mViewY1;
            mViewX2=mViewY2=-1;
            mMirrorX=mMirrorY=false;
            mWidth=0;mHeight=0;
            mSquare=false;
            mName=strdup("<view>");
            Recompute();
        }

        ~View()
        {
            if (mName)
                delete mName;
        }
        void SetName(const char* aName);
        const char* Name() { return mName; }        
        int ConvertXToView(double aX);
        int ConvertYToView(double aY);
        
        double ConvertViewToX(int aX);
        double ConvertViewToY(int aY);

        void SetView(int aX1,int aY1,int aX2,int aY2);
        void SetSquare(int aSquare) { mSquare=aSquare; }
        void Resize(int aWidth,int aHeight);

        double Zoom() { return mZoom; }
        double ScrollX() { return mScrollX; }
        double ScrollY() { return mScrollY; }
        int Height() { return mHeight; }
        int Width() { return mWidth; }
        
        void SetMirror(bool aMirrorX,bool aMirrorY) { mMirrorX=aMirrorX; mMirrorY=aMirrorY; }
        void SetZoom(double aZoom) { mZoom=aZoom; }
        void SetScroll(double aScrollX,double aScrollY) { mScrollX=aScrollX; mScrollY=aScrollY; }

        void Select(Display *aDisplay,GC aGC);

        bool Inside(int aX,int aY) { return ((aX>=mClipX1) && (aY<mClipX2) && (aY>=mClipY1) && (aY<mClipY2)); }
    };
    
class ViewerWindow
{
    protected:
		unsigned int mWidth, mHeight;
		Display *mDisplay;
		int mScreen;
		int mDepth;
        GC mGC;
		Window mWindow;
        Pixmap mBuffer;
        Drawable mHandle;
        class BasicActor* mActor;
        char* mTitle;
        bool mClosed;
        unsigned long mPenColor,mFillColor,mBackgroundColor;
        bool mFill;
        XFontStruct* mFont;
        void SelectPenColor();
        void SelectFillColor();
        void SelectFillColorBack();

        vector<View*> mViews;
        View *mCurrentView,*mDefaultView;        
    public:
        unsigned long mViewFlags;
        ViewerWindow()
        {
            mClosed=false;
            mBuffer=0;
            mActor=0;
            mDisplay=0;
            mHandle=0;
            mScreen=0;
            mDepth=0;
            mTitle=0;
            mGC=0;
            mPenColor=0x0;
            mFillColor=0xFFFFFF;
            mBackgroundColor=0xFFFFFF;
            mFill=false;
            mViews.push_back(mDefaultView=mCurrentView=new View);
            mFont=0;
            mViewFlags=0xFFFFFFFF;
        }
        // Open and close
		void Create(int width=800,int height=600);
		void Close();
        bool Closed() { return mClosed; }
        
        // Events
		void ProcessEvent(XEvent report);
		void BeginFrame();
        void EndFrame();
        
        //window properties
        void SetTitle(const char* aTitle);
        void SetBackgroundColor(unsigned long aRGB);

        View* CreateView(const char* aName);
        void SelectView(View* aView);
        View* GetView(unsigned int i) { if (i<mViews.size()) return mViews[i]; else return 0; }
        
        int Width() { return mWidth; }
        int Height() { return mHeight; }
        
        // Elemental draw functions
        
        void SetActor(class BasicActor* aActor) { mActor=aActor; }
        class BasicActor* Actor() { return mActor; }

        // Draw
        void SetPenColor(unsigned long aRGB) { mPenColor=aRGB; }
        void SetFillColor(unsigned long aRGB) { mFillColor=aRGB; }
        void SetFill(bool aFill) { mFill=aFill; }
        void SetFont(const char* aName,int aSize);
        
        void Clear();
             
        void Line(double aX1,double aY1,double aX2,double aY2,int aWidth=0,int aStyle=LineSolid);
        void Rectangle(double aX1,double aY1,double aX2,double aY2);
        void Plot(double aX,double aY);
        void Text(double aX,double aY,const string& aText);
        void Polygon(int aN,double* aVertices);
        void Marker(double aX,double aY,char aMarker,int aSize=2);
};


#endif

