#include "behaviourextinguish.h"
#include "../FireAgent.h"
#include "../Team.h"
#include "../SituationModel.h"
#include "../WorldModel.h"

class ExtinguishPositionFn : public SelectionFunction
{
protected:
    ImaginaryBuilding* target_building;
public:
    virtual bool Consider(ImaginaryObject * o) const
    {
        if (o==target_building)
        {
            return false;
        }
        ImaginaryBuilding* B=dynamic_cast<ImaginaryBuilding*>(o);
        float c=(config.max_extinguish_length())/1000.0;
        float x,y;
        
        if (B) // Don't stand in a burning building
        {
            if ((B->mHp<1.0) || (B->mFire!=0.0))
            {
                return false;
            }
            x=B->X()-target_building->X();
            y=B->Y()-target_building->Y();            
            return x*x + y*y < c*c;
        }
        // TODO check road for danger of becoming blocked by collapsing buildings?
        ImaginaryRoad* R=dynamic_cast<ImaginaryRoad*>(o);
        
        if (R) 
        {
            x=R->X()-target_building->X();
            y=R->Y()-target_building->Y();
            return x*x + y*y < c*c;
        }
        ImaginaryNode* N=dynamic_cast<ImaginaryNode*>(o);
        if (N)
        {
            x=N->X()-target_building->X();
            y=N->Y()-target_building->Y();
            return x*x+ y*y < c*c;
        }
        return false;
    }
    virtual float Value(ImaginaryObject *o) const
    {
        return 0.0;
    }
    ExtinguishPositionFn(ImaginaryBuilding* target) : SelectionFunction()
    {
        target_building=target;
    }
};    
    const float cIgniteDistance=30000.0;

class IsFireRiskNearObjectFn : public SelectionFunction
{
protected:
    ImaginaryBuilding* building;
public:
    IsFireRiskNearObjectFn(ImaginaryBuilding* o) : SelectionFunction()
    {
        building=o;
    }
    virtual bool Consider(ImaginaryObject* o) const
    {
        ImaginaryBuilding* B=dynamic_cast<ImaginaryBuilding*>(o);
        if (B)
        {
            if (building->mFire>=0.0)
            {
                float distance=o->DistanceTo(building);
                if (distance<cIgniteDistance) // TODO: calculate wind             {
                {
                    return building->mFire!=0.0;
                }
            }
        }
        return false;
    }
    virtual float Value(ImaginaryObject* o) const
    {
        return o->DistanceTo(building)/cIgniteDistance;
//        return 1.0;
    }
};

float BehaviourExtinguish::Priority(SituationModel* aModel,TeamMember* aActor)
{
    // TODO calc once for optimization and update on changes in neighbours and fieryness
    ImaginaryBuilding* B=aModel->GetBuilding(mTarget);
    if (B)
    {
        float score=B->mFire;
        float q=0.0f;
        ImaginaryObjects neighbours=aModel->FindObjects(IsFireRiskNearObjectFn(B));
        for(ImaginaryObjects::const_iterator I=neighbours.begin();I!=neighbours.end();I++)
        {
            if ((B=dynamic_cast<ImaginaryBuilding*>(*I)))
            {
                score+=1.0f-B->mFire;
                q+=1.0f;
            }
        }
        if (q>0.0f)
        {
            return (0.5f)+(score/(2.0f*q));
        }
        else
        {
            return 0.0f;
        }
    }
    else
    {
        return 0.0;
    }
}
float BehaviourExtinguish::Attainability(SituationModel* aModel,TeamMember* aActor)
{
    ImaginaryBuilding* B=aModel->GetBuilding(mTarget);

    if (B->mFire>0)
    {
        if ((aActor->IsFireBrigade()))
        {
            // note: this score is not how likely it is that the building will stop burning
            // only how likely it is that the exec can start extinguishing
            return aActor->Human()->mWater;
        }
    }
    return 0.0f;
}

SituationModel* BehaviourExtinguish::Predict(SituationModel* aSituation,TeamMember* aActor)
{
    SituationModel* out=Behaviour::Predict(aSituation,aActor);
    const ImaginaryObjects& extinguish_positions=aSituation->FindSortedObjects(ExtinguishPositionFn(aSituation->GetBuilding(mTarget)));
    if (extinguish_positions.Contains(aActor->Position()->ActualObject()))
    {
        ImaginaryBuilding* B=(ImaginaryBuilding*)out->Override(out->GetBuilding(mTarget));
        B->mWater+=0.1f;
        mPosition=0;
        return out;
    }
    else
    {
        for(ImaginaryObjects::const_iterator PI=extinguish_positions.begin();PI!=extinguish_positions.end();PI++)
        {
            SituationModel* out2=PredictTravel(out,aActor,(*PI)->ActualObject());
            if (out2)
            {
                return out2;
            }
        }
        return 0;
    }
}


void BehaviourExtinguish::Execute(BasicAgent* aExec)
{
    Debug(spam_behaviour_execution,"Executing %s by %s %04ld\n",ToString().c_str(),aExec->agentName(),aExec->ID());
    WorldModel* WM=aExec->model();

    if (mPosition)
    {
        Path* P=WM->FindPath(aExec->Self(),mPosition,Path::AdvancedTravelFn(aExec));
        if (P)
        {
            if (P->Length()==1)
            {
                FireAgent* FA=dynamic_cast<FireAgent*>(aExec);
                if (FA)
                {
                    FA->Extinguish(dynamic_cast<Building*>(mTarget));
                }
            }
            else
            {
                aExec->Move(P);
            }
        }
    }
    else
    {
         FireAgent* FA=dynamic_cast<FireAgent*>(aExec);
         if (FA)
         {
             FA->Extinguish(dynamic_cast<Building*>(mTarget));
         }
    }
}

string BehaviourExtinguish::ToString()
{
    return string("Extinguish "+IntToString(((Building*)mTarget)->id()));
}


