#include "behaviourfree.h"
#include "../PoliceAgent.h"
#include "../Team.h"
#include "../SituationModel.h"
#include "../WorldModel.h"
using namespace Rescue;

// TODO make road clear behaviour and derive Free, Sweep and Refuge from it

class ClearPositionFn : public SelectionFunction
{
protected:
    const Object* mTarget;
public:
    ClearPositionFn(const Object* aTarget) : SelectionFunction()
    {
        mTarget=aTarget;
    }
    virtual bool Consider(ImaginaryObject* o) const
    {
        if (o->ActualObject()==mTarget)
        {
            return true;
        }
        else
        {
            Node* N=dynamic_cast<Node*>(o->ActualObject());
            if (N)
            {
                for(Objects::const_iterator NI=N->neighbors().begin();NI!=N->neighbors().end();NI++)
                {
                    if (*NI==mTarget)
                    {
                        return true;
                    }
                }
            }
        }
        return false;
    }
    virtual float Value(ImaginaryObject *o) const
    {
        return 0.0f;
    }
};

class ReachableFreePositionFn : public SelectionFunction
{
protected:
    SituationModel* mModel;
    TeamMember* mActor;
    const Object* mTarget;
    const Path* mPath;
public:
    ReachableFreePositionFn(const Path* aPath,const Object* aTarget,SituationModel* aModel,TeamMember* aActor)
    {
        mPath=aPath;
        mActor=aActor;
        mTarget=aTarget;
        mModel=aModel;
    }
    virtual bool Consider(ImaginaryObject* o) const
    {
        if (mPath->Contains(o->ActualObject()))
        {
            Path* P=mModel->FindPath(mActor->Human(),o);
            if (P)
            {
                delete P;
                return  true;
            }
        }
        return false;
    }

    virtual float Value(ImaginaryObject* o) const
    {
        return 1.0f/(1.0f+log10(DistanceBetween(o->ActualObject(),mTarget)));
    }
};


float BehaviourFree::Priority(SituationModel* aModel,TeamMember* aActor)
{
    return 0.8+0.2/(1.0+log10(aActor->Human()->DistanceTo(aModel->GetHuman(mTarget))));
}

Rescue::Object* BehaviourFree::DetermineTarget(SituationModel* aModel,TeamMember* aActor)
{ 
    // find path from target TO me so target can get where I am
   
    // find closest nexus in target sector
    Object* targetnexus=0;
    float distance,bestDistance=MAXFLOAT;
    if (aModel->GetSector())
    {
        const objects_t& targetnexi=aModel->GetSector()->nexusPoints();
        for(objects_t::const_iterator NI=targetnexi.begin();NI!=targetnexi.end();NI++)
        {
            distance=DistanceBetween(*NI,mTarget);
            if (distance<bestDistance)
            {
                 bestDistance=distance;
                targetnexus=*NI;
            }
        }
        assert(targetnexus!=NULL);
    }
    
    Path *Route=aModel->FindPath(aActor->Human(),aModel->GetHuman(mTarget),false);
    if (targetnexus)
    {
        *Route+=*(aModel->FindPath(aModel->GetHuman(mTarget),aModel->GetNode(targetnexus),false));
    }

    const ImaginaryObjects& clear_positions=aModel->FindSortedObjects(ReachableFreePositionFn(Route,mTarget,aModel,aActor));
    delete Route;
    Route=0;
    for(ImaginaryObjects::const_iterator PI=clear_positions.begin();PI!=clear_positions.end();PI++)
    {
        Road* R=dynamic_cast<Road*>((*PI)->ActualObject());
        Node* N=dynamic_cast<Node*>((*PI)->ActualObject());
        if (R)
        {
            ImaginaryRoad* r=aModel->GetRoad(R);
            if (r->mBlock!=0.0)
            {
                Path* P=aModel->FindPath(aActor->Human(),r);
                if (P)   // check if we can get to this road
                {
                    delete P;
                    mPosition=R;
                    return R;
                }
            }
        }
        if (N)
        {
            for(Objects::const_iterator NI=N->neighbors().begin();NI!=N->neighbors().end();NI++)
            {
                R=dynamic_cast<Road*>(*NI);
                if (R)
                {
                    ImaginaryRoad* r=aModel->GetRoad(R);
                    if ((r) && (r->mBlock!=0.0))
                    {
                        Path* P=aModel->FindPath(aActor->Human(),*PI);
                        if (P)    // check if we can get to this node
                        {
                            delete P;
                            mPosition=N;
                            return R;
                        }
                    }
                }
            }
        }               
    }
    mPosition=0;
    return 0;
}

float BehaviourFree::Attainability(SituationModel* aModel,TeamMember* aActor)
{
    assert((aModel->GetHuman(mTarget)!=NULL) && (aModel->GetHuman(mTarget)->mPosition!=NULL));
    Object* pos=DetermineTarget(aModel,aActor);
    if ((pos) || (mPosition))
    {
        return 1.0f;
    }
    else
    {
        Path* P=aModel->FindPath(aActor->Human(),aModel->GetHuman(mTarget));
        if (P)
        {
            delete P;
            return 1.0;
        }
        else
        {
            return 0.0;
        }
    }
}

SituationModel* BehaviourFree::Predict(SituationModel* aSituation,TeamMember* aActor)
{
    SituationModel* out=Behaviour::Predict(aSituation,aActor);
    Rescue::Object* clear=DetermineTarget(aSituation,aActor);
    if (mPosition)
    {                 
        // Move or Act
        if ((mPosition==aActor->Position()->ActualObject()) || (clear==aActor->Position()->ActualObject()))
        {
            ImaginaryRoad* R=aSituation->GetRoad(clear);
            R=(ImaginaryRoad*)out->Override(R);
            R->mBlock-=1.0f;
            mClear=clear;
            Debug(spam_behaviour_selection,"Predicting free to clear %ld\n",mClear->id());
            return out;
        }
        else
        {
            mClear=0;
            Debug(spam_behaviour_selection,"Predicting free to move to %ld and clear %ld\n",mPosition->id(),clear->id());
            return PredictMove(out,aActor,mPosition);
        }
    }
    else
    {
        Debug(spam_behaviour_selection,"Predicting free to travel to target %ld\n",mTarget->id());
        mPosition=mTarget;
        mClear=0;
        return PredictTravel(out,aActor,mTarget);
    }
}


void BehaviourFree::Execute(BasicAgent* aExec)
{
    Debug(spam_behaviour_execution,"Executing %s by %s %04ld\n",ToString().c_str(),aExec->agentName(),aExec->ID());
    WorldModel* WM=aExec->model();
    if (mClear)
    {
        PoliceAgent* PA=dynamic_cast<PoliceAgent*>(aExec);
        if (PA)
        {
            PA->ClearRoad(dynamic_cast<Road*>(mClear));
        }
    }
    else
    {
        if ((mPosition) && (mPosition!=aExec->Position()))
        {
            Path* P=WM->FindPath(aExec->Self(),mPosition,Path::AdvancedTravelFn(aExec));
            if (P)
            {
                aExec->Move(P);
                delete P;
            }
        }
    }
}



string BehaviourFree::ToString()
{
    return string("Free "+IntToString(mTarget->id()));
}

