#include "behavioursweeppath.h"
#include "../PoliceAgent.h"
#include "../Team.h"
#include "../SituationModel.h"
#include "../WorldModel.h"
using namespace Rescue;

class BlockedRoadFn : public SelectionFunction
{
protected:
    const Path* mPath;
public:
    BlockedRoadFn(const Path* aPath) : SelectionFunction()
    {
        mPath=aPath;
    }
    virtual bool Consider(ImaginaryObject* o) const
    {
        ImaginaryRoad* R=dynamic_cast<ImaginaryRoad*>(o);
        if (R)
        {
            Road* r=(Road*)R->ActualObject();
            return (R->mBlock>0.0f) &&
            ((mPath->Contains(r->head())) || (mPath->Contains(r->tail())));
        }
        return false;
    }
    virtual float Value(ImaginaryObject *o) const
    {
        ImaginaryRoad* R=dynamic_cast<ImaginaryRoad*>(o);
        if (R)
        {
            return R->mBlock;
        }
        return 0.0f;
    }
};

class ClearPositionFn : public SelectionFunction
{
protected:
    const Object* mTarget;
public:
    ClearPositionFn(const Object* aTarget) : SelectionFunction()
    {
        mTarget=aTarget;
    }
    virtual bool Consider(ImaginaryObject* o) const
    {
        if (o->ActualObject()==mTarget)
        {
            return true;
        }
        else
        {
            Node* N=dynamic_cast<Node*>(o->ActualObject());
            if (N)
            {
                for(Objects::const_iterator NI=N->neighbors().begin();NI!=N->neighbors().end();NI++)
                {
                    if (*NI==mTarget)
                    {
                        return true;
                    }
                }
            }
        }
        return false;
    }
    virtual float Value(ImaginaryObject *o) const
    {
        return 0.0f;
    }
};

class ReachablePositionFn : public SelectionFunction
{
protected:
    SituationModel* mModel;
    TeamMember* mActor;
    const Path* mPath;
public:
    ReachablePositionFn(const Path* aPath,SituationModel* aModel,TeamMember* aActor)
    {
        mPath=aPath;
        mActor=aActor;
        mModel=aModel;
    }
    virtual bool Consider(ImaginaryObject* o) const
    {
        return (mPath->Contains(o->ActualObject()));
    }

    virtual float Value(ImaginaryObject* o) const
    {
        Path* P=mModel->FindPath(mActor->Human(),o);
        if (P)
        {
            float out=1.0f/(1.0f+log(1.0f+(float)P->Length()));
            delete P;
            return out;
        }
        else
        {
            return 0.0f;
        }
    }
};


float BehaviourSweepPath::Priority(SituationModel* aModel,TeamMember* aActor)
{

    // TODO calc once for optimization and update on changes in neighbours and fieryness
    float block=0.0f;
    float q=1.0f;
    const Objects &objs=mPath->GetObjects();
    
    for(Objects::const_iterator PI=objs.begin();PI!=objs.end();PI++)
    {
        ImaginaryRoad* R=aModel->GetRoad(*PI);
        if (R)
        {
            block+=R->mBlock;
            q+=2.0f;
        }
    }
    return 0.5+(block/q);
}

Rescue::Object* BehaviourSweepPath::DetermineTarget(SituationModel* aModel,TeamMember* aActor)
{
    const ImaginaryObjects& clear_positions=aModel->FindObjects(ReachablePositionFn(mPath,aModel,aActor));
    for(ImaginaryObjects::const_iterator PI=clear_positions.begin();PI!=clear_positions.end();PI++)
    {
        Road* R=dynamic_cast<Road*>((*PI)->ActualObject());
        Node* N=dynamic_cast<Node*>((*PI)->ActualObject());
        if (R)
        {
            ImaginaryRoad* r=aModel->GetRoad(R);
            if (r->mBlock!=0.0)
            {
                Path* P=aModel->FindPath(aActor->Human(),r);
                if (P)   // check if we can get to this road
                {
                    delete P;
                    mPosition=R;
                    return R;
                }
            }
        }
        if (N)
        {
            for(Objects::const_iterator NI=N->neighbors().begin();NI!=N->neighbors().end();NI++)
            {
                R=dynamic_cast<Road*>(*NI);
                if (R)
                {
                    ImaginaryRoad* r=aModel->GetRoad(R);
                    if ((r) && (r->mBlock!=0.0))
                    {
                        Path* P=aModel->FindPath(aActor->Human(),*PI);
                        if (P)    // check if we can get to this node
                        {
                            delete P;
                            mPosition=N;
                            return R;
                        }
                    }
                }
            }
        }               
    }
    mPosition=0;
    return 0;
}

float BehaviourSweepPath::Attainability(SituationModel* aModel,TeamMember* aActor)
{
    DetermineTarget(aModel,aActor);
    return mTarget!=0;
}

SituationModel* BehaviourSweepPath::Predict(SituationModel* aSituation,TeamMember* aActor)
{
    SituationModel* out=Behaviour::Predict(aSituation,aActor);
    Rescue::Object* clear=DetermineTarget(aSituation,aActor);
    if (mPosition)
    {
        // Move or Act
        if ((mPosition==aActor->Position()->ActualObject()) || (clear==aActor->Position()->ActualObject()))
        {
            ImaginaryRoad* R=aSituation->GetRoad(clear);
            R=(ImaginaryRoad*)out->Override(R);
            R->mBlock-=1.0f;
            mClear=clear;
            return out;
        }
        else
        {
            mClear=0;
            return PredictMove(out,aActor,mPosition);
        }
    }
    else
    {
        return 0;
    }
}


void BehaviourSweepPath::Execute(BasicAgent* aExec)
{
    Debug(spam_behaviour_execution,"Executing %s by %s %04ld\n",ToString().c_str(),aExec->agentName(),aExec->ID());
    WorldModel* WM=aExec->model();
    Debug(spam_behaviour_execution,"Position = %ld Clear = %ld\n",(mPosition)?mPosition->id():0,(mClear)?mClear->id():0);

    if ((mPosition) && (mPosition!=aExec->Position()))
    {
        Path* P=WM->FindPath(aExec->Self(),mPosition,Path::AdvancedTravelFn(aExec));
        if (P)
        {
            if (P->Length()==1)
            {
                PoliceAgent* PA=dynamic_cast<PoliceAgent*>(aExec);
                if (PA)
                {
                    PA->ClearRoad(dynamic_cast<Road*>(mClear));
                }
            }
            else
            {
                aExec->Move(P);
            }
        }
    }
    else
    {
         PoliceAgent* PA=dynamic_cast<PoliceAgent*>(aExec);
         if (PA)
         {
             PA->ClearRoad(dynamic_cast<Road*>(mClear));
         }
    }
}



string BehaviourSweepPath::ToString()
{
    return string("Sweep path "+mPath->ToString());
}
