// Controller.hxx
//
// This is a base class to derive controllers from.  It provides support
//  for an ObjectPool as memory.  Derived controllers need only implement
//  the act function, and optionally sense, hear, and init.
//
// Multiple controllers automatically share the same socket.  They can also
//  share memory by passing the same memory pointer in the constructor.
//
// A number of convenience functions are provided for putting together
//  commands to send to the server.  packet_* provides low-level 
//  support for commands.  The act_* commands provide means for
//  sending complete commands.
//

#ifndef _Controller_hxx
#define _Controller_hxx

#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"

namespace Rescue {

class Controller
{
public:
  static void setup_connection(char *host, int port);

  Controller(int type, ObjectPool *memory);
  virtual ~Controller();

  void getSelf(Input * input);
  bool connectOK(Input * input,int tmp_id);
  bool connect();

  static void go();
  static void start();
  static bool loop();
  static void end();

protected:

  //
  // Basic Functions
  //

  virtual void init() {};
  virtual void sense_change(Object *o) {};
  virtual void sense_object(S32 time, Object *o, Input &input);
  virtual void hear(Object *sender, const char *msg) {};
  virtual void act() { send(); }

  enum ActPolicy { ACT_ONCE_WHEN_IDLE, ACT_ONCE_PER_CYCLE, ACT_ONCE_PER_SENSE, ACT_WHEN_IDLE, 
                   ACT_LATER, ACT_NEVER };
  ActPolicy m_act_policy; 

  void send();

  virtual void sense(S32 time, Input &input);

  //
  // Memory functions.
  //

  ObjectPool *m_memory;

  template <class T>
  T *lookup(Id id) { return dynamic_cast<T *>(m_memory->get(id)); }

  template <class T>
  T *is_a(Object *o) { return dynamic_cast<T *>(o); }
  
  S32 m_time;
  S32 m_id;
  MovingObject *m_self;

  //
  // Action Functions
  //

  void act_move(Object **path); // path must be a NULL terminated array.
  void act_open(Road *target);
  void act_rescue(Humanoid *target);
  void act_load(Humanoid *target);
  void act_unload();

  void act_say(Object *target, const char *msg);
  void act_tell(Object *target, const char *msg);

  void act_extinguish_simple(Building *target);
  void act_say_list(Object **targets, const char *msg);
  void act_tell_list(Object **targets, const char *msg);

  void act_extinguish(); // Unimplemented

  //
  // Implementation Details.  These are only made available to derived
  // classes to maximize extensibility.
  //

  int m_type;

  // A mapping of IDs to controllers.
  typedef std::map<Id, AutoPtr<Controller> > IdMap;
  static IdMap m_controllers;
  // adjusted by UvaRescue2003: moved these two to here (they were locals in go)
  // the last_controller signifies the position in the control loop
  static IdMap::iterator last_controller;
  // old_time is the time of the previous iteration in the control loop
  static S32 old_time;

  // A flag set if the Controller's act function should be called at 
  //  the next available opportunity.
  int m_call_act;

  // The connection is shared among all controllers.  It therefore
  // must be initialized before the connect() is called.  This is done
  // by calling setup_connection().
  static Address m_address;
  static LongUDPSocket m_socket;

  // Convenience functions for putting together commands to the
  // kernel.  These mainly handle setting the size of the packet after
  // it's been assembled.  They also handle the socket connection.
  Output m_output;
  Output::Cursor *m_output_base;

  void packet_start(Output &output, Header header);
  void packet_end(Output &output);
  void packet_send(Output &output);

private:
  // Used for performance.  Only downloads initial GIS data once.
  static Input obj_input;
};
  
}

#endif

