// Memory.cxx

#include "Memory.hxx"

using namespace Rescue;
 
Object *Memory::find(const ObjectFn &query, FunctionOrder order)
{
  Object *best_o = NULL;
  double best_score = 0.0;

  for(Objects::iterator it = m_objects.begin(); it != m_objects.end(); it++) {
    if (!query.consider(*it)) continue;

    double score = query.f(*it);
    if (!best_o || 
	(order == FUNCTION_MAXIMIZE && score > best_score) || 
	(order == FUNCTION_MINIMIZE && score < best_score)) {
      best_o = (*it);
      best_score = score;
    }
  }

  return best_o;
}

Object **Memory::find_list(const ObjectFn &query, 
			   double threshold, FunctionOrder order)
{
  Objects os;

  for(Objects::iterator it = m_objects.begin(); it != m_objects.end(); it++) {
    if (!query.consider(*it)) continue;

    double score = query.f(*it);
    if ((order == FUNCTION_MAXIMIZE && score >= threshold) || 
	(order == FUNCTION_MINIMIZE && score <= threshold))
      os.push_front(*it);

  }

  Object **rv = new (Object*)[os.size() + 1];
  int i = 0;

  while(!os.empty()) { rv[i++] = os.front(); os.pop_front(); }
  rv[i] = NULL;

  return rv;
  
}

// This structure is the ordering relation for the A* search, used in
//  the method find_shortest_path.
struct SearchCompare { 
  std::map<Id, double> *p;
  bool operator()(const Id &a, const Id &b) { 
    if ((*p)[a] == (*p)[b]) return a < b;
    else return (*p)[a] < (*p)[b]; 
  }
};

// Uses A* search on the GIS road network.
//
// Takes an edge cost function and a heuristic function.  Defaults
//  for these are road length and straight line distance, respectively.
//
Object **Memory::find_path(const Location &origin, 
			   const ObjectFn &destination,
			   const ObjectFn &heuristic,
			   const ObjectPairFn &edge_cost,
			   bool overlap_check) 
{
  std::map<Id, Id> search_path;
  std::map<Id, double> search_cost;
  std::map<Id, double> search_priority;
  SearchCompare search_compare = { &search_priority };
  std::set<Id,SearchCompare> search_queue(search_compare);

  Id origin_id = origin.m_o->id();

  // SETTING UP THE SEARCH
  //
  // This is some ugliness to deal with searches that originate
  //  from the middle of a road.  Basically, the two endpoints of
  //  the road need to be added into the queue to begin.
  //
  // If the origin is just a node/building then we just use it for
  //  start of the queue.
  //

  if (origin.m_o->type() == TYPE_ROAD) {
    Road *o = (Road *) origin.m_o;
    Object *head = o->head();
    Object *tail = o->tail();

    search_path[origin_id] = origin_id;
    search_cost[origin_id] = 0;

    if ((edge_cost.consider(o, tail) || origin.m_dist <= 0.0) &&
	heuristic.consider(head)) {
      Id id = head->id();
      search_path[id] = origin_id;
      search_cost[id] = (edge_cost.f(origin.m_o, tail) * 
			 origin.m_dist / (double) o->length());
      search_queue.insert(id);
      search_priority[id] = heuristic.f(head);
    }

    if ((edge_cost.consider(o, head) || origin.m_dist >= o->length()) &&
	heuristic.consider(tail)) {
      Id id = tail->id();
      search_path[id] = origin_id;
      search_cost[id] = (edge_cost.f(origin.m_o, head) * 
			 (1.0 - origin.m_dist / (double) o->length()));
      search_queue.insert(id);
      search_priority[id] = heuristic.f(tail);
    }

  } else {
    search_path[origin_id] = origin_id;
    search_cost[origin_id] = 0;
    search_priority[origin_id] = 0;
    search_queue.insert(origin_id);
  }

  //
  // THE SEARCH
  //

  Object *destination_found = NULL;

#ifdef TIMING
  int nodes_visited = 0;
  struct itimerval itimer = {{100000, 0}, {100000, 0}};
  setitimer(ITIMER_REAL, &itimer, NULL);
#endif
  
  while(1) {
    if (search_queue.empty()) break;
    Id o_id = *search_queue.begin(); search_queue.erase(search_queue.begin());
    MotionlessObject *o = get_typed<MotionlessObject>(o_id);

#ifdef TIMING
    nodes_visited++;
#endif

    if (destination.consider(o)) { destination_found = o; break; }

    // Add all neighbors to the queue.
    for(Objects::const_iterator it = o->neighbors().begin();
	it != o->neighbors().end(); it++) {
      const MotionlessObject *n = dynamic_cast<MotionlessObject *>(*it);
      Id n_id = n->id();
      
      if (n == o || !heuristic.consider(n) ||
	  !edge_cost.consider(n, o)) continue;

     // UvA-Rescue-C2003: overlap check      

     if (overlap_check) {
	bool overlap=false;
        for(Id i = o_id; i != origin_id; i = search_path[i]) {
            if (i==n_id) {
               overlap=true;
               break;
            }
        }
        if (overlap)
          continue;
      }

      double dist = search_cost[o_id] + edge_cost.f(n, o);

      if (search_cost.find(n_id) != search_cost.end() &&
	  search_cost[n_id] <= dist) continue;

      search_cost[n_id] = dist;
      search_priority[n_id] = dist + heuristic.f(n);
      search_path[n_id] = o_id;
      search_queue.insert(n_id);
    }
  }

#ifdef TIMING
  getitimer(ITIMER_REAL, &itimer);
  double secs = 100000.0 - (itimer.it_value.tv_sec + 
			itimer.it_value.tv_usec / 1000000.0);
  itimer.it_value.tv_sec = 0;
  itimer.it_value.tv_usec = 0;
  setitimer(ITIMER_REAL, &itimer, NULL);

  if (!destination_found) 
    fprintf(stderr, "Visited %d nodes in %g secs, and found no path.\n",
	    nodes_visited, secs);
#endif

  //
  // THE RESULTS
  //

  if (!destination_found) return NULL;

  // Determine the length of the path.
  Id dest_id = destination_found->id();
  int path_length = 1;
  for(Id i = dest_id; i != origin_id; path_length++, i = search_path[i]);

#ifdef TIMING
  fprintf(stderr, "Visited %d nodes in %g secs, to find path of length %d.\n",
	  nodes_visited, secs, path_length);
#endif


  // Store the path in a newly allocated array.
  Object **path = new (Object*) [path_length]; path_length--;
  path[path_length--] = NULL;
  for(Id i = dest_id; i != origin_id; path_length--, i = search_path[i])
    path[path_length] = get(i);

  return path;
}

bool Memory::verify_path(Object **path, const ObjectPairFn &edge_cost) 
{
  // This currently checks that the edge_cost is valid for each edge.
  for( ; *path; path++) {
    MotionlessObject *o = dynamic_cast<MotionlessObject*>(*path);
    MotionlessObject *no = dynamic_cast<MotionlessObject*>(*(path + 1));

    if (no) {
      if (!edge_cost.consider(no, o)) break;

      Objects::const_iterator it;

      for(it = o->neighbors().begin(); 
	  it != o->neighbors().end(); it++)
	if ((*it) == no) break;

      if (it == o->neighbors().end()) break;
    }
  }

  if (!*path || !*(path + 1)) return true;
  else return false;
}

Object **Memory::update_path(Object **path, Object *here,
			     const ObjectFn &destination,
			     const ObjectPairFn &edge_cost)
{
  if (!path) return NULL;
  
  Object **sub_path = path_subpath(path, here);
  Object *dest = path_destination(path);

  /* Skip the first object in the list... this is just 'here'. */
  if (sub_path) sub_path++;

  if (!sub_path || 
      !destination.consider(dest) || 
      !verify_path(sub_path, edge_cost)) {
    delete [] path;
    return NULL;
  }

  /* The sub_path may be the entire path. */
  if (sub_path == path) return sub_path;

  /* Copy sub_path to beginning of path.  Or is it faster to use memmove? */
  int i;
  for(i=0; sub_path[i]; i++) path[i] = sub_path[i];
  path[i] = 0;
 
  return path;
}


