// Controller.cxx
//

#include "Controller.hxx"

#undef VERBOSE

using namespace Rescue;

LongUDPSocket Controller::m_socket("");
Address Controller::m_address;
Controller::IdMap Controller::m_controllers;
Input *Controller::m_initial_input = NULL;

Controller::Controller(int type, ObjectPool *memory)
{
  m_type = type;
  m_time = -1;
  m_id = 0;

  m_memory = memory;

  m_self = NULL;
  m_output_base = NULL;

  m_act_policy = ACT_ONCE_WHEN_IDLE;
  m_call_act = false;


  m_output.clear();
}

Controller::~Controller()
{
}

void Controller::setup_connection(char *host, int port)
{
  Output output;

  m_address = Address(host, port);
  output.clear();
  output.put(HEADER_NULL);
  m_socket.send(output, m_address, -1);
}

bool Controller::connect()
{
  static int tmp_id = 1;
  Output output;
  Input *input=new Input();

  fprintf(stderr, "Connecting... ");

  int version = 0;

  // This is an oddity with shared memory.  We only need the initial
  // GIS information once.  By specifying a version number of 1 we
  // skip the retransfer of the same information.  So we need to check
  // whether this memory has already be initialized.
  // if (!m_memory->objects().empty()) version = 1;
  if (m_initial_input) version = 1;

  output.clear();
  packet_start(output, AK_CONNECT);
  output.put((S32) ++tmp_id);      // Temporary ID
  output.put((S32) version);       // Version?
  output.put((S32) m_type);        // Agent Type
  packet_end(output);
  packet_send(output);

  while(1) {
    m_socket.receive(*input);

    // Read packet
    Header header = input->get();
    Input::Cursor start = input->cursor();
    S32 size = input->get();

    if (header == KA_CONNECT_ERROR) {
      fprintf(stderr, "FAILED! (%ld)\n", header);
      input->setCursor(start); input->skip(size);
      return 0;
    } else if (header != KA_CONNECT_OK || tmp_id != input->get()) {
      input->setCursor(start); input->skip(size);
      continue;
    } else {

      m_time = 0;
      m_id = input->get();

      fprintf(stderr, "OK (ID = %ld)\n", m_id); 

      if (!m_initial_input)
      {
        m_initial_input=input;
      }
      else
      {

      }
      Input::Cursor c=m_initial_input->cursor();
      m_memory->restructure(m_time,*m_initial_input);
      m_initial_input->setCursor(c);
      if (m_initial_input!=input)
      {
        m_memory->restructure(m_time,*input);
        delete input;
      }
/*
      if (m_initial_input) {
Input::Cursor c = m_initial_input->cursor();
	m_memory->restructure(m_time, *m_initial_input);
	m_initial_input->setCursor(c);
      } else {
	m_initial_input = new Input(input);
      }
      m_memory->restructure(m_time, input);

      input->setCursor(start);
      input->skip(size);
*/

      // Send acknowledgement
      output.clear();
      packet_start(output, AK_ACKNOWLEDGE);
      output.put(m_id);
      packet_end(output);
      packet_send(output);
  
      // Add to list of controllers.
      m_controllers[m_id] = std::AutoPtr<Controller>(this);

      // Do any controller specific initilization.
      init();

      return 1;
    }
  }

}
  
void Controller::go()
{
  IdMap::iterator last_controller = m_controllers.begin();

  Input input;
#if VERBOSE
  S32 old_time = -1;
#endif

  for(;;) {
    // If the socket's empty then we're idling and so an agent should have
    //  their act() function called.
    if (m_socket.empty()) {

      // Find next controller that needs to act. 
      IdMap::iterator i = last_controller;

      for(i++; ; i++) {
	if (i == m_controllers.end()) i = m_controllers.begin();
	if (i == last_controller || (*i).second.get()->m_call_act) break;
      }
	
      if ((*i).second.get()->m_call_act) {
	Controller *c = (*i).second.get();

	c->act(); c->send();
	if (c->m_act_policy == ACT_ONCE_WHEN_IDLE) c->m_call_act = false;

	last_controller = i;
      } else {
	// Since no controller needs to act and there's no input, we
	//  can just wait for some input.
	vector<const LongUDPSocket *> v;
	v.push_back(&m_socket);
	LongUDPSocket::wait(v, 1000);
      }

      continue;
    }

    m_socket.receive(input);

    // Examine header.  
    Header header = input.get();
    Input::Cursor start = input.cursor(); // Beginning of input.
    S32 size = input.get();

    switch(header) {
    case KA_SENSE:
    case KA_HEAR: {
      Id id = input.get();
      Controller *receiver = m_controllers[id].get();
      
      if (!receiver) {
	fprintf(stderr, "Unknown ID (%ld). Skipping...\n", id);
	input.setCursor(start);
	input.skip(size);
	continue;
      }

      if (header == KA_SENSE) {
	S32 time = input.get();

#if VERBOSE
	if (time > old_time) {
	  fprintf(stderr, "Time = %ld.\n", time);
	  old_time = time;
	}
#endif

	if (receiver->m_act_policy != ACT_NEVER)
	  receiver->sense(time, input);

	switch (receiver->m_act_policy) {
	case ACT_ONCE_PER_SENSE:
	  receiver->act(); receiver->send(); break;
	case ACT_WHEN_IDLE:
	case ACT_ONCE_WHEN_IDLE:
	  receiver->m_call_act = true; break;
	case ACT_LATER:
	case ACT_NEVER:
	  break;
	}

      } else {
	Id sender = input.get();
	S32 size = (input.get() + 2) & ~3;
	Byte msg[size + 1];

	input.get(size, msg); msg[size] = '\0';
	receiver->hear(receiver->lookup<Object>(sender), (const char *) msg);
      }

      input.setCursor(start);
      input.skip(size);

    } break;

    default:
      fprintf(stderr, "Unknown header (%ld).  Skipping...\n", header);
      input.setCursor(start);
      input.skip(size);
      continue;
    }

  }
}

void Controller::packet_start(Output &output, Header header)
{
  output.put(header);
if (m_output_base) delete m_output_base;
  m_output_base = new Output::Cursor(output.put(~(S32) 0));
}

void Controller::packet_end(Output &output)
{
  output.setSize(*m_output_base);
}

void Controller::packet_send(Output &output)
{
  output.put(HEADER_NULL);
  m_socket.send(output, m_address, m_time);
}

void Controller::sense(S32 time, Input &input)
{
  TypeId type;

  m_time = time;
  while((type = input.get()) != TYPE_NULL)
    sense_object(time, m_memory->get(input.get()), input);

  if (!m_self) m_self = lookup<MovingObject>(m_id);
}

void Controller::sense_object(S32 time, Object *o, Input &input)
{
  o->inputAllProperty(time, *m_memory, input);
  sense_change(o);
}

void Controller::send()
{
  packet_send(m_output);
  m_output.clear();
}

void Controller::act_move(Object **path)
{
  packet_start(m_output, AK_MOVE);
  m_output.put(m_id);

  // For paths of length one, we have to be workaround a bug currently
  //  in the traffic simulator.  If the destination is not a building, we
  //  must first specify a street, the agent's current node, then the building.
  if (path[0] && !path[1] && is_a<Node>(m_self->position())) {
    Node *n = (Node *) m_self->position();
    
    for(Objects::const_iterator i = n->edges().begin(); 
	i != n->edges().end(); i++)
      if (is_a<Road>(*i)) {
	m_output.put((*i)->id());
	m_output.put(n->id());
	break;
      }

    m_output.put(path[0]->id());
  } else {
    for(int i=0; path[i]; i++) m_output.put(path[i]->id());
  }

  m_output.put((Id) 0);
  packet_end(m_output);

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Move: %ld -> %ld.\n", m_id, 
	  m_self->position()->id(), path[0]->id());
#endif

}

void Controller::act_open( Road *target)
{
  packet_start(m_output, AK_CLEAR);
  m_output.put(m_id);
  m_output.put(target->id());
  packet_end(m_output);

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Open: %ld (%ld%%).\n", 
	  m_id, target->id(), (100 * target->block()) / target->width());
#endif
}

void Controller::act_rescue(Humanoid *target)
{
  packet_start(m_output, AK_RESCUE);
  m_output.put(m_id);
  m_output.put(target->id());
  packet_end(m_output);

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Rescue: %ld.\n", m_id, target->id());
#endif
}

void Controller::act_load(Humanoid *target)
{
  packet_start(m_output, AK_LOAD);
  m_output.put(m_id);
  m_output.put(target->id());
  packet_end(m_output);

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Load: %ld.\n", m_id, target->id());
#endif
}


void Controller::act_unload()
{
  packet_start(m_output, AK_UNLOAD);
  m_output.put(m_id);
  packet_end(m_output);

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Unload.\n", m_id);
#endif
}


void Controller::act_say(Object *target, const char *msg)
{
  packet_start(m_output, AK_SAY);
  m_output.put(m_id);
  m_output.putString(msg);
  m_output.put(target->id());
  packet_end(m_output);

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Say: %ld, \"%s\".\n", m_id, target->id(), msg);
#endif
}

void Controller::act_tell(Object *target, const char *msg)
{
  packet_start(m_output, AK_TELL);
  m_output.put(m_id);
  m_output.putString(msg);
  m_output.put(target->id());
  packet_end(m_output);

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Tell: %ld, \"%s\".\n", m_id, target->id(), msg);
#endif
}

void Controller::act_say_list(Object **targets, const char *msg)
{
  for(int i=0; targets[i]; i++) {
    if (targets[i] == m_self) continue;

    packet_start(m_output, AK_SAY);
    m_output.put(m_id);
    m_output.putString(msg);
    m_output.put(targets[i]->id());
    packet_end(m_output);
  }

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Say: LIST, \"%s\".\n", m_id, msg);
#endif
}

void Controller::act_tell_list(Object **targets, const char *msg)
{
  for(int i=0; targets[i]; i++) {
    if (targets[i] == m_self) continue;

    packet_start(m_output, AK_TELL);
    m_output.put(m_id);
    m_output.putString(msg);
    m_output.put(targets[i]->id());
    packet_end(m_output);
  }

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Tell: LIST, \"%s\".\n", m_id, msg);
#endif
}

void Controller::act_extinguish_simple(Building *target)
{
  packet_start(m_output, AK_EXTINGUISH);
  m_output.put(m_id);
  m_output.put(target->id());
  m_output.put(0); // Direction (??)
  m_output.put(target->x());
  m_output.put(target->y());
  m_output.put(1000); // Amount of water (per minute)
  m_output.put((Id) 0); // ? 
  packet_end(m_output);

#ifdef VERBOSE
  fprintf(stderr, "[%ld] Extinguish: %ld.\n", m_id, target->id());
#endif
}

